@@ -506,7 +506,8 @@ tibetan_engine_shape (PangoEngineShape *engine,
 
   fc_font = PANGO_FC_FONT (font);
   face = pango_fc_font_lock_face (fc_font);
-  g_return_if_fail (face != NULL);
+  if (!face)
+    return;
 
   buffer = pango_ot_buffer_new (fc_font);
 
