@@ -58,6 +58,33 @@ set_color (PangoCairoRenderer *crenderer,
 			  color->blue / 65535.);
 }
 
+static void
+_pango_cairo_renderer_draw_box_glyph (PangoCairoRenderer *crenderer,
+				      PangoGlyphInfo     *gi,
+				      double              cx,
+				      double              cy)
+{
+  double temp_x, temp_y;
+
+  cairo_save (crenderer->cr);
+  cairo_get_current_point (crenderer->cr, &temp_x, &temp_y);
+
+  cairo_rectangle (crenderer->cr,
+		   cx + 1.5,
+		   cy + 1.5 - PANGO_UNKNOWN_GLYPH_HEIGHT,
+		   (double)gi->geometry.width / PANGO_SCALE - 3.0,
+		   PANGO_UNKNOWN_GLYPH_HEIGHT - 3.0);
+
+  if (!crenderer->do_path)
+    {
+      cairo_set_line_width (crenderer->cr, 1.0);
+      cairo_stroke (crenderer->cr);
+    }
+
+  cairo_move_to (crenderer->cr, temp_x, temp_y);
+  cairo_restore (crenderer->cr);
+}
+
 static void
 _pango_cairo_renderer_draw_unknown_glyph (PangoCairoRenderer *crenderer,
 					  PangoFont          *font,
@@ -66,59 +93,65 @@ _pango_cairo_renderer_draw_unknown_glyph (PangoCairoRenderer *crenderer,
 					  double              cy)
 {
   char buf[7];
-  double ys[2];
-  double xs[3];
+  double x0, y0;
   int row, col;
-  int cols;
+  int rows, cols;
   char hexbox_string[2] = {0, 0};
   double temp_x, temp_y;
   PangoCairoHexBoxInfo *hbi;
   gunichar ch;
 
-  hbi = _pango_cairo_get_hex_box_info ((PangoCairoFont *)font);      
-
-  ch = gi->glyph & ~PANGO_CAIRO_UNKNOWN_FLAG;
-
-  cols = ch > 0xffff ? 3 : 2;
-  g_snprintf (buf, sizeof(buf), cols == 2 ? "%04X" : "%06X", ch);
-
-  ys[1] = cy + hbi->box_descent - hbi->pad * 2;
-  ys[0] = ys[1] - hbi->digit_height - hbi->pad;
-
-  xs[0] = cx + hbi->pad * 3.0;
-  xs[1] = xs[0] + hbi->digit_width + hbi->pad;
-  xs[2] = xs[1] + hbi->digit_width + hbi->pad;
-
   cairo_save (crenderer->cr);
   cairo_get_current_point (crenderer->cr, &temp_x, &temp_y);
 
+  hbi = _pango_cairo_font_get_hex_box_info ((PangoCairoFont *)font);      
+
+  if (!hbi || !_pango_cairo_font_install (PANGO_CAIRO_FONT (hbi->font), crenderer->cr))
+    {
+      _pango_cairo_renderer_draw_box_glyph (crenderer, gi, cx, cy);
+      goto done;
+    }
+
+  ch = gi->glyph & ~PANGO_GLYPH_UNKNOWN_FLAG;
+
+  rows = hbi->rows;
+  cols = (ch > 0xffff ? 6 : 4) / rows;
+  g_snprintf (buf, sizeof(buf), (ch > 0xffff) ? "%06X" : "%04X", ch);
+
   cairo_rectangle (crenderer->cr,
-		   cx + hbi->pad * 1.5,
-		   cy + hbi->box_descent - hbi->pad * 0.5,
-		   (double)gi->geometry.width / PANGO_SCALE - 3 * hbi->pad,
-		   -(hbi->box_height - hbi->pad));
+		   cx + hbi->pad_x * 1.5,
+		   cy + hbi->box_descent - hbi->box_height + hbi->pad_y * 0.5,
+		   (double)gi->geometry.width / PANGO_SCALE - 3 * hbi->pad_x,
+		   (hbi->box_height - hbi->pad_y));
 
   if (!crenderer->do_path)
     {
-      cairo_save (crenderer->cr);
-      cairo_set_line_width (crenderer->cr, hbi->pad);
+      cairo_set_line_width (crenderer->cr, hbi->line_width);
       cairo_stroke (crenderer->cr);
-      cairo_restore (crenderer->cr);
     }
 
-  _pango_cairo_font_install (PANGO_CAIRO_FONT (hbi->font), crenderer->cr);
-  for (row = 0; row < 2; row++)
+  x0 = cx + hbi->pad_x * 3.0;
+  y0 = cy + hbi->box_descent - hbi->pad_y * 2;
+
+  for (row = 0; row < rows; row++)
+    {
+      double y = y0 - (rows - 1 - row) * (hbi->digit_height + hbi->pad_y);
       for (col = 0; col < cols; col++)
 	{
+	  double x = x0 + col * (hbi->digit_width + hbi->pad_x);
+
+	  cairo_move_to (crenderer->cr, x, y);
+
 	  hexbox_string[0] = buf[row * cols + col];
-	  cairo_move_to (crenderer->cr, xs[col], ys[row]);
 
 	  if (crenderer->do_path)
 	      cairo_text_path (crenderer->cr, hexbox_string);
 	  else
 	      cairo_show_text (crenderer->cr, hexbox_string);
 	}
+    }
 
+done:
   cairo_move_to (crenderer->cr, temp_x, temp_y);
   cairo_restore (crenderer->cr);
 }
@@ -147,6 +180,27 @@ pango_cairo_renderer_draw_glyphs (PangoRenderer     *renderer,
       set_color (crenderer, PANGO_RENDER_PART_FOREGROUND);
     }
 
+  if (!_pango_cairo_font_install (PANGO_CAIRO_FONT (font), crenderer->cr))
+    {
+      for (i = 0; i < glyphs->num_glyphs; i++)
+	{
+	  PangoGlyphInfo *gi = &glyphs->glyphs[i];
+
+	  if (gi->glyph != PANGO_GLYPH_EMPTY)
+	    {
+	      double cx = crenderer->x_offset +
+			  (double)(x + x_position + gi->geometry.x_offset) / PANGO_SCALE;
+	      double cy = crenderer->y_offset +
+			  (double)(y + gi->geometry.y_offset) / PANGO_SCALE;
+
+	      _pango_cairo_renderer_draw_unknown_glyph (crenderer, font, gi, cx, cy);
+	    }	  
+	  x_position += gi->geometry.width;
+	}
+
+      goto done;
+    }
+
   if (glyphs->num_glyphs > MAX_STACK)
     cairo_glyphs = g_new (cairo_glyph_t, glyphs->num_glyphs);
   else
@@ -157,12 +211,14 @@ pango_cairo_renderer_draw_glyphs (PangoRenderer     *renderer,
     {
       PangoGlyphInfo *gi = &glyphs->glyphs[i];
 
-      if (gi->glyph)
+      if (gi->glyph != PANGO_GLYPH_EMPTY)
         {
-          double cx = crenderer->x_offset + (double)(x + x_position + gi->geometry.x_offset) / PANGO_SCALE;
-          double cy = crenderer->y_offset + (double)(y + gi->geometry.y_offset) / PANGO_SCALE;
+          double cx = crenderer->x_offset +
+		      (double)(x + x_position + gi->geometry.x_offset) / PANGO_SCALE;
+          double cy = crenderer->y_offset +
+		      (double)(y + gi->geometry.y_offset) / PANGO_SCALE;
 
-          if (gi->glyph & PANGO_CAIRO_UNKNOWN_FLAG)
+          if (gi->glyph & PANGO_GLYPH_UNKNOWN_FLAG)
 	    _pango_cairo_renderer_draw_unknown_glyph (crenderer, font, gi, cx, cy);
           else
             {
@@ -175,8 +231,6 @@ pango_cairo_renderer_draw_glyphs (PangoRenderer     *renderer,
       x_position += gi->geometry.width;
     }
 
-  _pango_cairo_font_install (PANGO_CAIRO_FONT (font), crenderer->cr);
-  
   if (crenderer->do_path)
     cairo_glyph_path (crenderer->cr, cairo_glyphs, count);
   else
@@ -185,6 +239,7 @@ pango_cairo_renderer_draw_glyphs (PangoRenderer     *renderer,
   if (glyphs->num_glyphs > MAX_STACK)
     g_free (cairo_glyphs);
 
+done:
   if (!crenderer->do_path)
     cairo_restore (crenderer->cr);
   
@@ -212,11 +267,13 @@ pango_cairo_renderer_draw_rectangle (PangoRenderer     *renderer,
 		   crenderer->x_offset + (double)x / PANGO_SCALE,
 		   crenderer->y_offset + (double)y / PANGO_SCALE, 
 		   (double)width / PANGO_SCALE, (double)height / PANGO_SCALE);
+
   if (!crenderer->do_path)
-    cairo_fill (crenderer->cr);
-  
-  if (!crenderer->do_path)
-    cairo_restore (crenderer->cr);
+    {
+      cairo_fill (crenderer->cr);
+    
+      cairo_restore (crenderer->cr);
+    }
 }
 
 /* Draws an error underline that looks like one of:
@@ -362,13 +419,19 @@ pango_cairo_show_glyph_string (cairo_t          *cr,
   PangoFontMap *fontmap;
   PangoCairoRenderer *crenderer;
   PangoRenderer *renderer;
+  gboolean unref_renderer = FALSE;
 
   g_return_if_fail (cr != NULL);
-  g_return_if_fail (PANGO_IS_CAIRO_FONT (font));
   g_return_if_fail (glyphs != NULL);
 
   fontmap = pango_font_get_font_map (font);
   renderer = _pango_cairo_font_map_get_renderer (PANGO_CAIRO_FONT_MAP (fontmap));
+  if (G_UNLIKELY (!renderer))
+    {
+      renderer = g_object_new (PANGO_TYPE_CAIRO_RENDERER, NULL);
+      unref_renderer = TRUE;
+    }
+
   crenderer = PANGO_CAIRO_RENDERER (renderer);
 
   cairo_save (cr);
@@ -388,9 +451,14 @@ pango_cairo_show_glyph_string (cairo_t          *cr,
 
   pango_renderer_deactivate (renderer);
   
-  crenderer->cr = NULL;
-  crenderer->x_offset = 0.;
-  crenderer->y_offset = 0.;
+  if (G_UNLIKELY (unref_renderer))
+    g_object_unref (renderer);
+  else
+    {
+      crenderer->cr = NULL;
+      crenderer->x_offset = 0.;
+      crenderer->y_offset = 0.;
+    }
   
   cairo_restore (cr);
 }
@@ -501,13 +569,19 @@ pango_cairo_glyph_string_path (cairo_t          *cr,
   PangoFontMap *fontmap;
   PangoCairoRenderer *crenderer;
   PangoRenderer *renderer;
+  gboolean unref_renderer = FALSE;
 
   g_return_if_fail (cr != NULL);
-  g_return_if_fail (PANGO_IS_CAIRO_FONT (font));
   g_return_if_fail (glyphs != NULL);
 
   fontmap = pango_font_get_font_map (font);
   renderer = _pango_cairo_font_map_get_renderer (PANGO_CAIRO_FONT_MAP (fontmap));
+  if (G_UNLIKELY (!renderer))
+    {
+      renderer = g_object_new (PANGO_TYPE_CAIRO_RENDERER, NULL);
+      unref_renderer = TRUE;
+    }
+
   crenderer = PANGO_CAIRO_RENDERER (renderer);
 
   crenderer->cr = cr;
@@ -516,10 +590,15 @@ pango_cairo_glyph_string_path (cairo_t          *cr,
   
   pango_renderer_draw_glyphs (renderer, font, glyphs, 0, 0);
   
-  crenderer->cr = NULL;
-  crenderer->do_path = FALSE;
-  crenderer->x_offset = 0.;
-  crenderer->y_offset = 0.;
+  if (G_UNLIKELY (unref_renderer))
+    g_object_unref (renderer);
+  else
+    {
+      crenderer->cr = NULL;
+      crenderer->do_path = FALSE;
+      crenderer->x_offset = 0.;
+      crenderer->y_offset = 0.;
+    }
   
   cairo_set_font_face (cr, NULL);
 }
